/* Implementation of ON THE EFFECTIVENESS OF WEIGHT-ENCODED NEURAL IMPLICIT 3D SHAPES (https://arxiv.org/pdf/2009.09808.pdf).
* The idea is to represent a neural implicit by its weights and biases. The architecture is N fully connected layers, with hidden layer
* size equals to H. The non-linearity of the hidden layers is a ReLU and the output layer is activated by a tanh.
*/

// 7296 = 3 * N for input layer, N * N * (H - 1) for hidden layers, and H for output layer.
RWStructuredBuffer<float> weights : register(u0);

// 257 = H * N + 1 biases
RWStructuredBuffer<float> biases : register(u1);

const int N = 8; // Number of layers
const int H = 32; // Hidden size

float leakyReLU(float val)
{
	return max(0.f, val) + 0.1 * min(0.f, val);
}

// Hidden layer computation.
void matvecmul(in out float buffer[64], bool relu, int M, int K, int weight_idx, int input_idx, int bias_idx, int output_idx)
{
	for (int i = 0; i < M; ++i)
	{
		float val = 0.f;
		for (int j = 0; j < K; ++j)
		{
			val += weights[weight_idx + i * K + j] * buffer[input_idx + j];
		}
		val += biases[bias_idx + i];
		buffer[output_idx + i] = (relu) ? leakyReLU(val) : val;
	}
}

float computeSDF(float3 pos)
{
	float3 val = float3(1, 0, 0);

	int M = H, K = 3;

	bool relu = true;

	// This buffer stores data for intermediate computations: input and output for a layer. Size = 2 * H = 64.
	float buffer[64];

	buffer[0] = pos.x;
	buffer[1] = pos.y;
	buffer[2] = pos.z;
    
    int weight_idx = 0;
    int bias_idx = 0;
	int output_idx = 0;
    for (int l = 0; l < N + 1; ++l)
	{
		int input_idx = M * (l % 2);
		output_idx = M * ((l + 1) % 2);
		if (l == N)
		{
			M = 1;
			relu = false;
		}

		matvecmul(buffer, relu, M, K, weight_idx, input_idx, bias_idx, output_idx);

		weight_idx += M * K;
		bias_idx += M;

		K = M;
	}

	return buffer[output_idx];
	//float result = tanh(buffer[output_idx]);
	//return result;
}

float4 main(float4 fragCoord : SV_POSITION) : SV_TARGET
{
	float sdf = computeSDF(float3(10.f, 0.f, 0.f));

    return float4(sdf, 0.f, 0.f, 1.f);
}
